/*
** Module   : DataModelWorker.java
** Abstract : Pattern worker with specific knowledge of the data model
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050426   @21002 Created initial version. Modeled after
**                           uast.JavaPatternWorker.
** 002 ECF 20050526   @21303 Simplified worker library methods. The new
**                           expression engine implementation enables us
**                           to reference objects directly in expressions,
**                           so much of the worker library implementation
**                           which jumped through hoops to use AST ids in
**                           lieu of object references was eliminated.
** 003 ECF 20050614   @21536 Added capability to sort class properties.
**                           The order associated with the peer field's
**                           "order" value is used to sort into ascending
**                           order. Also removed visitAst to prevent root
**                           node creation when visiting a previously
**                           generated AST. This is now done from rulesets
**                           using the createRootAst user function.
** 004 ECF 20050714   @21703 Added capability to sort class ASTs within a
**                           data model root AST. Classes are sorted by
**                           dependency, with dependent classes ordered
**                           after those classes upon which they are
**                           dependent. Also removed unused methods.
** 005 ECF 20050720   @21776 Modified sortProperties method. Remove
**                           temporary "order" annotation from properties
**                           after the sort completes.
** 006 ECF 20050805   @22012 Rewrote algorithm to sort class ASTs by
**                           dependency. The old approach did not always
**                           properly resolve dependencies.
** 007 ECF 20051005   @22961 Change to P2O filename generation. Now ".p2o"
**                           replaces the ".schema" extension, instead of
**                           being appended to it.
** 008 ECF 20051027   @23215 Added user functions. normalizeUniqueIndexes
**                           calculates minimal set of unique constraints
**                           for a list of indexes. nonRedundantIndexes
**                           calculates the non-redundant subset of a list
**                           of indexes.
** 009 ECF 20060504   @25998 Fixed property sorting. Collections.sort()
**                           was not working properly in the temp table
**                           case. Changed collection type from TreeSet to
**                           ArrayList.
** 010 GES 20070629   @34339 Removed terse flag from AST persistence so
**                           that we retain line/column info in our ASTs.
** 011 ECF 20080730   @39270 Reduced memory footprint. Implemented finish
**                           method to clean up resources. Removed current
**                           instance variable, which held an unused, hard
**                           reference to a potentially very large AST.
** 012 GES 20090515   @42219 Moved to AstManager from AstRegistry/AstPersister.
** 013 GES 20090518   @42395 Import change.
** 014 ECF 20121220          Fixed sortClasses to accommodate sequence ASTs,
**                           which are siblings of class ASTs. This method
**                           assumed class ASTs were the only children of the
**                           data model root AST.
** 015 ECF 20130216          Fixed null handling in PropertyComparator.
**/

package com.goldencode.p2j.schema;

import java.io.*;
import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.persist.*;

/**
 * A pattern worker that provides services to support the conversion of
 * Progress schema data into a relational data model AST.
 */
public final class DataModelWorker
extends AbstractConversionWorker
implements DataModelTokenTypes
{
   /** File extension appended to schema filenames */
   private static final String P2O_POSTFIX = ".p2o";
   
   /** Root of the target tree currently being processed. */
   private DataModelAst root = null;
   
   /**
    * Worker routine to centralize creation of new data model ASTs.  All ID
    * setup and parent linkages are established if the parent AST is passed
    * as a parameter, otherwise only an unidentified, standalone AST is
    * created. If a parent AST is passed, a resolver instance should also
    * be provided. 
    *
    * @param    type
    *           The token type for the data model AST.
    * @param    text
    *           The text for the AST.
    * @param    parent
    *           The parent AST node or <code>null</code>.
    * @param    resolver
    *           The current instance of the {@link AstSymbolResolver} in use
    *           or <code>null</code>.
    *
    * @return   The AST that was created.
    */
   public static DataModelAst createAst(int type,
                                        String text,
                                        Aast parent,
                                        AstSymbolResolver resolver)
   {
      DataModelAst ast = new DataModelAst();
      initializeAst(ast, type, text, parent, resolver);
      
      return ast;
   }
   
   /**
    * Worker routine for peer AST creation.  All ID setup and parent linkages
    * are established and bidirectional peer ID annotations are written
    * in source and target nodes.
    *
    * @param    type
    *           The token type for the AST.
    * @param    text
    *           The text for the AST.
    * @param    parent
    *           The parent AST node.
    * @param    resolver
    *           The current instance of the {@link AstSymbolResolver} in use.
    *
    * @return   The AST that was created.
    */
   public static DataModelAst createPeerAst(Aast source,
                                            int type,
                                            String text,
                                            Aast parent,
                                            AstSymbolResolver resolver)
   {
      DataModelAst target = createAst(type, text, parent, resolver);
      
      // save our cross-references
      crossReferencePeerAsts(source, target);
            
      return target;
   }
   
   /**
    * Default constructor which initializes libraries.
    */
   public DataModelWorker()
   {
      super();
      
      setLibrary(new Library());
   }
   
   /**
    * Worker routine to centralize creation of a new data model AST that is
    * a root node.  A new file ID is registered in the {@link AstManager}
    * and set into the AST.
    *
    * @param   peerRoot
    *          Root node of Progress schema AST currently being processed.
    *
    * @return  The root AST that was created.
    *
    * @throws  ConfigurationException
    *          if P2J environment is not properly initialized.
    */
   public Aast createRootAst(Aast peerRoot)
   throws ConfigurationException
   {
      // reset the current/root AST to null
      root = null;
      resolver.setTargetRootAst(null);
      
      // Create a new target tree
      root = createAst(DATA_MODEL, "data model", null, null);
      String filename = generateFilename(peerRoot);
      File file = Configuration.toFile(filename);
      initializeRootAst(root, file, resolver);
      resolver.setTargetRootAst(root);
      
      return root;
   }
   
   /**
    * Resolve <code>constant</code> to a literal which can be compiled into
    * expressions.
    *
    * @param   constant
    *          Constant indicating a data model token name.
    *
    * @return  Token type associated with the token name.
    */
   public Object resolveConstant(String constant)
   {
      return DataModelAst.resolveConstant(constant);
   }
   
   /**
    * Hook to provide termination processing for a pattern worker.
    */
   public void finish()
   {
      root = null;
      
      super.finish();
   }
   
   /**
    * Generate a unique data model AST persistence filename based on the
    * current source AST filename. The source file name's extension is
    * removed and replaced with ".p2o".
    * 
    * @param   ast
    *          Current source AST.
    *
    * @return   A unique filename based on the source file root name with
    *          ".p2o" appended.
    */
   private String generateFilename(Aast ast)
   {
      String filename = ast.getFilename();
      filename = filename.substring(0, filename.lastIndexOf("."));
      StringBuffer buf = new StringBuffer(filename);
      buf.append(P2O_POSTFIX);
      
      return buf.toString();
   }
   
   /**
    * Helper to create, edit, move and delete data model ASTs.
    */
   public class Library
   {
      /**
       * Create a new data model AST that is the root for the data model
       * hierarchy. This can later be retrieved with the {@link #getAstRoot}
       * user function.
       *
       * @param   peerRoot
       *          Root node of Progress schema AST currently being processed.
       *
       * @return  The root AST that was created.
       *
       * @throws  ConfigurationException
       *          if P2J environment is not properly initialized.
       */
      public Aast createRootAst(Aast peerRoot)
      throws ConfigurationException
      {
         return DataModelWorker.this.createRootAst(peerRoot);
      }
      
      /**
       * General purpose data model AST creation function which can create a
       * <code>DataModelAst</code> node using the given token type and text. 
       * The created node will then be attached to the specified parent. This
       * is used in cases where the conversion is simple enough to drive
       * it completely from the rule set rather than from a custom worker.
       *
       * @param    type
       *           The token type for the data model AST.
       * @param    text
       *           The text for the AST.
       * @param    parent
       *           The ID of the parent AST node.
       *
       * @return   The unique ID if the AST was created successfully or
       *           <code>null</code> if the parent ID is invalid, if the type
       *           is not valid or on any unexpected error during AST 
       *           creation.
       */
      public Aast createAst(int type, String text, Aast parent)
      {
         return DataModelWorker.createAst(type, text, parent, resolver);
      }
      
      /**
       * General purpose data model AST creation function which can create a
       * <code>DataModelAst</code> node using the given token type and text. 
       * The created node will then be attached to the specified parent. Most
       * importantly, on success the ID of the generated AST node will be
       * stored in the current source AST node's "peerid" annotation. This
       * provides a standard mechanism to duplicate the structure of a source
       * AST in the target data model AST.  The source AST's ID will also
       * be stored in the target AST's "peerid" annotation which allows 
       * subsequent bidirectional referencing between these peers. This
       * is used in cases where the conversion is simple enough to drive
       * it completely from the rule set rather than from a custom worker.
       *
       * @param    type
       *           The token type for the data model AST.
       * @param    text
       *           The text for the AST.
       * @param    parent
       *           The ID of the parent AST node.
       *
       * @return   The unique ID if the AST was created successfully or
       *           <code>null</code> if the parent ID is invalid, if the type
       *           is not valid or on any unexpected error during AST 
       *           creation.
       */
      public Aast createPeerAst(int type, String text, Aast parent)
      {
         return DataModelWorker.createPeerAst(getCopy(),
                                              type,
                                              text,
                                              parent,
                                              resolver);
      }
      
      /**
       * Return the AST node that is the root of the data model AST
       * hierarchy.
       *
       * @return  Data model AST root node.
       */
      public Aast getAstRoot()
      {
         return root;
      }
      
      /**
       * Common case for data model AST creation which creates a 
       * <code>DataModelAst</code> node using the passed token type and text 
       * and then attaches that node to the parent specified by the "peerid"
       * annotation of the source AST's parent node. This greatly simplifies
       * creation of entire hierarchies of structure that match between the
       * source and target ASTs.  This uses the services of {@link
       * #createPeerAst(int,String,Aast)}.
       *
       * @param    type
       *           The token type for the data model AST.
       * @param    text
       *           The text for the AST.
       *
       * @return   The unique ID if the AST was created successfully or
       *           <code>null</code> if the parent ID is invalid, if the type
       *           is not valid or on any unexpected error during AST 
       *           creation.
       */
      public Aast createPeerAst(int type, String text)
      {
         Long id = (Long) getCopy().getParent().getAnnotation(PEER_ID);
         if (id == null)
         {
            return null;
         }
         Aast parentPeer = resolver.getAst(id);
         
         return createPeerAst(type, text, parentPeer);
      }
      
      /**
       * Sorts <code>PROPERTY</code> children ASTs of <code>parent</code>
       * in place into ascending order, according to the numeric value
       * specified by the "order" annotation. Sibling ASTs of other types
       * are left in place.
       *
       * @param   parent
       *          Parent AST of type <code>CLASS</code>.
       */
      public void sortProperties(Aast parent)
      {
         List list = new ArrayList();
         Aast child = parent.getImmediateChild(PROPERTY, null);
         while (child != null)
         {
            list.add(child);
            child = parent.getImmediateChild(PROPERTY, child);
         }
         
         Collections.sort(list, new PropertyComparator());
         
         // Remove temporary "order" annotation.
         Iterator iter = list.iterator();
         while (iter.hasNext())
         {
            child = (Aast) iter.next();
            child.removeAnnotation("order");
         }
         
         // If list contains 1 or fewer entries, no need to sort.
         if (list.size() <= 1)
         {
            return;
         }
         
         Aast preProp = null;
         Aast postProp = null;
         child = (Aast) parent.getFirstChild();
         
         // Detect first non-PROPERTY child before PROPERTY ASTs.
         while (child != null)
         {
            if (child.getType() == PROPERTY)
            {
               break;
            }
            preProp = child;
            child = (Aast) child.getNextSibling();
         }
         
         // Detect first non-PROPERTY child after PROPERTY ASTs.
         while (child != null)
         {
            if (child.getType() != PROPERTY)
            {
               postProp = child;
               break;
            }
            child = (Aast) child.getNextSibling();
         }
         
         // Determine first and last PROPERTY ASTs, chaining together
         // siblings in between.
         Aast first = null;
         Aast last = null;
         iter = list.iterator();
         for (int i = 0; iter.hasNext(); i++)
         {
            child = (Aast) iter.next();
            if (i == 0)
            {
               first = child;
            }
            
            if (last != null)
            {
               last.setNextSibling(child);
            }
            
            last = child;
         }
         
         // Link first PROPERTY AST to head of list or to parent.
         if (preProp == null)
         {
            parent.setFirstChild(first);
         }
         else
         {
            preProp.setNextSibling(first);
         }
         
         // Link last PROPERTY AST to tail of list (may be null).
         last.setNextSibling(postProp);
      }
      
      /**
       * Sorts <code>CLASS</code> children ASTs of <code>parent</code>
       * in place into ascending order, according to the association
       * dependencies of each class. Classes are sorted after classes upon
       * which they have dependencies. Within dependency levels, they are
       * sorted alphabetically by class name.
       *
       * @param   parent
       *          Parent AST of type <code>DATA_MODEL</code>.
       */
      public void sortClasses(Aast parent)
      {
         SortedSet<Aast> set = new TreeSet<Aast>(new ClassComparator());
         Aast child = parent.getImmediateChild(CLASS, null);
         while (child != null)
         {
            set.add(child);
            child = parent.getImmediateChild(CLASS, child);
         }
         
         // If set contains 1 or fewer entries, no need to sort.
         if (set.size() <= 1)
         {
            return;
         }
         
         // Remove all CLASS nodes from their original locations.
         Iterator<Aast> iter = set.iterator();
         while (iter.hasNext())
         {
            Aast cls = iter.next();
            cls.remove();
         }
         
         // Classes are now sorted into 2 broad categories:  independent and
         // dependent. They are also sorted alphabetically within those
         // categories. We must now sort out the dependent classes to account
         // for the actual dependencies.
         List<Aast> list = new LinkedList<Aast>();
         iter = set.iterator();
         while (iter.hasNext())
         {
            Aast cls = iter.next();
            if (cls.annotationSize("dependency") == 0)
            {
               // Because of the first level sort, all independent classes
               // will always be added before any dependent classes.
               list.add(cls);
               continue;
            }
            
            // Walk backwards through the list until we find a class upon
            // which the current class is dependent. Add the current class
            // just after it. Not the most efficient algorithm, but it will
            // do...
            int pos = list.size();
            ListIterator<Aast> subIter = list.listIterator(pos);
            while (subIter.hasPrevious())
            {
               Aast next = subIter.previous();
               if (isXDependentUponY(cls, next))
               {
                  break;
               }
               pos--;
            }
            
            list.add(pos, cls);
         }
         
         /*
         // Debug output.
         AstSymbolResolver resolver = AstSymbolResolver.getResolver();
         iter = list.iterator();
         while (iter.hasNext())
         {
            Aast next = iter.next();
            System.out.println(next.getText());
            int size = next.annotationSize("dependency");
            for (int i = 0; i < size; i++)
            {
               Long depId = (Long) next.getAnnotation("dependency", i);
               System.out.println("   " + resolver.getAst(depId).getText());
            }
         }
         */
         
         // Re-attach to tree in sorted order. Remove temporary list of
         // dependencies from dependent class AST nodes.
         iter = list.iterator();
         while (iter.hasNext())
         {
            child = iter.next();
            child.removeAnnotation("dependency");
            parent.addChild(child);
         }
      }
      
      /**
       * Save the current target data model AST to file if the tree is not
       * empty (the root has at least 1 child).  If this AST is empty, it
       * will be removed from the {@link AstManager} which will essentially
       * disable further use of this AST.
       *
       * @return   <code>true</code> if the tree is persisted, 
       *           <code>false</code> if the tree is empty or on any failure  
       *           during persistence.
       */
      public boolean persistTarget()
      {
         String filename = generateFilename(getSource());
         
         // bypass persistence if there is an empty tree
         if (root.getNumImmediateChildren() == 0)
         {
            try
            {
               String normal = Configuration.normalizeFilename(filename);
               
               // no need to maintain this value in the registry
               AstManager.get().removeTree(normal);
            }
            catch (Exception exc)
            {
               exc.printStackTrace();
            }
            
            return false;
         }
         
         try
         {
            filename = Configuration.normalizeFilename(filename);
            
            // read the existing target tree
            AstManager.get().saveTree(root, filename, false);
         }
         catch (Exception exc)
         {
            // some failure, we can't persist!
            exc.printStackTrace();
            
            return false;
         }
         
         return true;
      }
      
      /**
       * Convenience method to expose to rulesets the {@link P2JIndex} static
       * method of the same name.
       *
       * @see  P2JIndex#normalizeUniqueIndexes
       */
      public List normalizeUniqueIndexes(List indexes)
      {
         return P2JIndex.normalizeUniqueIndexes(indexes);
      }
      
      /**
       * Convenience method to expose to rulesets the {@link P2JIndex} static
       * method of the same name.
       *
       * @see  P2JIndex#nonRedundantIndexes
       */
      public List nonRedundantIndexes(List indexes)
      {
         return P2JIndex.nonRedundantIndexes(indexes);
      }
   }
   
   /**
    * Tests one class AST for dependence upon another. The dependency test
    * includes the tested class <code>x</code>, as well as the entire tree
    * of its dependencies. In other words, test whether <code>x</code> <i>or
    * any class upon which it is dependent</i> is dependent upon
    * <code>y</code>.
    *
    * @param   x
    *          Class AST node whose dependencies are being checked.
    * @param   y
    *          Class AST node against which to test <code>x</code> for
    *          dependence.
    *
    * @return  <code>true</code> if <code>x</code> (or any of its dependent
    *          classes) is dependent upon <code>y</code>.
    */
   private boolean isXDependentUponY(Aast x, Aast y)
   {
      Long testId = y.getId();
      AstSymbolResolver resolver = AstSymbolResolver.getResolver();
      int len = x.annotationSize("dependency");
      for (int i = 0; i < len; i++)
      {
         Long depId = (Long) x.getAnnotation("dependency", i);
         Aast depAst = resolver.getAst(depId);
         if (depId.equals(testId) || isXDependentUponY(depAst, y))
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Comparator implementation which sorts ASTs which have an "order"
    * annotation into ascending order, according to the numeric value of
    * that attribute.
    */
   private static class PropertyComparator
   implements Comparator<Aast>
   {
      /**
       * Compare two ASTs and return a value indicating their relative sort
       * order. It is assumed that each object parameter represents an
       * annotated AST, and that each such AST has an "order" annotation.
       *
       * @param   a1
       *          First AST.
       * @param   a2
       *          Second AST.
       *
       * @return  A negative number if <code>o1</code> &lt; <code>o2</code>,
       *          a negative number if <code>o1</code> &gt; <code>o2</code>,
       *          else <code>0</code>.
       */
      public int compare(Aast a1, Aast a2)
      {
         Long order1 = (Long) a1.getAnnotation("order");
         Long order2 = (Long) a2.getAnnotation("order");
         
         if (order1 != null && order2 != null)
         {
            return (int) (order1.longValue() - order2.longValue());
         }
         
         if (order1 != null)
         {
            return 1;
         }
         
         if (order2 != null)
         {
            return -1;
         }
         
         return 0;
      }
   }
   
   /**
    * Comparator implementation which sorts ASTs which represent data model
    * object (DMO) classes. These classes are object representations of
    * relational tables. Foreign key relations create dependencies between
    * these tables. This comparator sorts classes such that independent
    * classes appear first in the sequence, followed by classes which are
    * dependent upon them.
    * <p>
    * This sort only separates completely independent classes from those
    * which have dependencies on other classes, but it <i>does not</i> sort
    * out the dependencies themselves. A second pass sort is used for that.
    */
   private static class ClassComparator
   implements Comparator<Aast>
   {
      /**
       * Compare two ASTs and return a value indicating their relative sort
       * order. It is assumed that each object parameter represents an
       * annotated AST, and that each such AST represents a data model object
       * class.
       *
       * @param   a1
       *          First AST.
       * @param   a2
       *          Second AST.
       *
       * @return  A negative number if <code>o1</code> &lt; <code>o2</code>,
       *          a negative number if <code>o1</code> &gt; <code>o2</code>,
       *          else <code>0</code>.
       */
      public int compare(Aast a1, Aast a2)
      {
         // Independent classes sort earlier than dependent ones.
         int len1 = a1.annotationSize("dependency");
         int len2 = a2.annotationSize("dependency");
         if (len1 == 0 && len2 > 0)
         {
            return -1;
         }
         if (len2 == 0 && len1 > 0)
         {
            return 1;
         }
         
         // No sorting could be determined by dependency, so just sort
         // in ascending alpha order by name.
         return (a1.getText().compareTo(a2.getText()));
      }
   }
}
